/*
 * MidpGameMapplet.java
 *
 * Created on 27 de diciembre de 2003, 16:46
 */

package org.jtgl.micro;

import java.util.Random;

import org.jtgl.core.*;
import org.jtgl.image.*;
import org.jtgl.input.*;
import org.jtgl.micro.*;

/**
 * GameMapplet is a base class for a gaming Mapplet.
 * @author  Manuel Polo (manuel_polo at yahoo dot es)
 */
public abstract class GameMapplet extends Mapplet implements Runnable{
    public final static int REFRESH_MANUAL  = -1;
    public final static int REFRESH_AUTO    = 0;
    public final static int MIN_FPS         = 1;
    public final static int MAX_FPS         = 1000;    
    private Thread gameThread;
    private KeyInputControl keyInputControl;
    private JTGLGraphics gc;    
    private Random random;
    private boolean initialized;
    private boolean runningGameLoop;
    private boolean refreshGraphics;
    private int refreshType;
    private int fps;
    private int refreshInterval;
    private long lastRefreshInterval;
    private boolean measureFPS;
    private long frameCounter;
    private long framesSum;   
    private long frameMCounter;
    private long frameTime,maxFrameTime,minFrameTime;
    private long realtimeFPS;
    private long realtimeMFPS;
    
    public GameMapplet() {        
        setFps(50);
    }    
  
    
    public void start(){
        gc = getGraphics();
        if(gc == null)
            return ;
        reset();
        if(gameThread == null){
            gameThread = new Thread(this);
            gameThread.setPriority(Thread.MAX_PRIORITY-1);
        }
        gameThread.start();                
    }
    
    public void pause() {
        reset();
    }    
    
    /**
     * Dirty method for complete reinitialization of game
     */    
    public void reset(){
        if(Thread.currentThread() == gameThread)
            return;
        debugMsg("Resetting...");
        gameThread = null;  
        try{
            while(isRunning()){
                debugMsg("Waiting for game thread to end...");
                delay(5);
            }
        }catch(Exception ex){  
            debugMsg("Exception: "+ex);
        }
        debugMsg("Resetting DONE!");
        initialized = false;
    }
    
    /** Kill Mapplet */    
    public void kill(boolean force) throws JTGLException{
        reset();
    }       
    public void setRefreshType(int type){
        refreshType = type;
    }
    
    public int getRefreshType(){
        return refreshType;
    }
    
    public void refreshGraphics(){
        refreshGraphics = true;
    }
    
    
    public final void setFps(int fps){
        this.fps = fps < MIN_FPS ? MIN_FPS : fps > MAX_FPS ? MAX_FPS : fps;
        refreshInterval = 1000/fps;
        debugMsg("refreshInterval = "+refreshInterval);
    }
    
    public final int getFps(){
        return fps;
    }     
    
    
    public void enableFPSMeasure(boolean enable){
        measureFPS = enable;
        if(!enable){
            realtimeFPS = realtimeMFPS = 0;
        }
    }
    
    public final long getFrameCounter(){
        return frameCounter;
    }
    
    public final long getRealFPS(){
        return realtimeFPS;
    }
    
    public final long getRealMFPS(){        
        return frameMCounter > 0 ? realtimeMFPS = framesSum/frameMCounter : 0;
    }    
    
    public final long getFrameTime(){
        return frameTime;
    }
    
    public final long getMaxFrameTime(){
        return maxFrameTime;
    }

    public final long getMinFrameTime(){
        return minFrameTime;
    }
   
    public int getKeys(){
        if(keyInputControl == null)
            keyInputControl = (KeyInputControl)getInputControl(KeyInputControl.CONTROL_CLASSNAME);        
        return keyInputControl == null ? KeyInputControl.KEY_NONE : keyInputControl.getKeys();
    }        
        
   
    
    public int getRandom(int to) {
        return getRandom(0,to);
    }
    
    public int getRandom(int from, int to) {
        if(random == null)
            random = new Random();
        if(from > to) {
            int save = from;
            from = to;
            to = save;
        }
        int range = to - from + 1;
        // compute a fraction of the range, 0 <= frac < range
        int frac = (random.nextInt() >>> 1) % range;
        return (frac + from);        
    }

    public boolean isRunning(){
        return runningGameLoop;
    }
    
    public void run() {
        if(Thread.currentThread() != gameThread)
            return;        
        if(!initialized){
            try {
                debugMsg("Init Game...");
                initGame();
                initialized = true;
            }            
            catch(Exception ex){
                ex.printStackTrace();
                handleException(ex);
                notifyKill();  
                return;
            }        
        }
        debugMsg("Started game loop!");
        long startFrame;               
        maxFrameTime = frameTime = startFrame = 0L;
        minFrameTime = 1<<30;        
        frameCounter = frameMCounter = 0;
        refreshGraphics();
        runningGameLoop = true;
        while(gameThread != null){                            
            frameCounter++;
            startFrame = System.currentTimeMillis();
            //#if debug            
            try {             
            //#endif
                processGameLogic();
                if(refreshType == REFRESH_AUTO)
                    drawGameScene();                        
                else
                    if(refreshGraphics){ //REFRESH_MANUAL
                        drawGameScene();
                        refreshGraphics = false;
                    }
            //#if debug
                
            }catch(Throwable t){
                t.printStackTrace();
            }                
            //#endif
            frameTime = System.currentTimeMillis() - startFrame;
            if(frameTime > maxFrameTime) maxFrameTime = frameTime; //Remember maximum frame time            
            if(frameTime < minFrameTime) minFrameTime = frameTime; //Remember minimum frame time            
            /////////// DELAY ////////////////            
            
            if(frameTime < refreshInterval){
                try {                         
                    Thread.sleep(refreshInterval-frameTime);                    
                }catch(Exception ex){}
            }else {                                
                Thread.yield();
            }             
            /////////// END DELAY ////////////
            
            if(measureFPS && frameTime > 0){
               frameMCounter++; //Count measures
               realtimeFPS = 1000 / frameTime;            
               //framesSum += realtimeFPS;//Measure median of realtime cycle (excluding refreshInterval delay)
               framesSum += 1000 / (System.currentTimeMillis() - startFrame);//Measure median of complete cycle (including refreshInterval delay)
            }            
        }
        runningGameLoop = false;
        debugMsg("End of game thread!");
    }      
    
    protected void initGame() throws Exception{
    }
   
    protected abstract void processGameLogic();
    
    //TODO: Change to receive the correct JTGLGraphics context
    protected abstract void drawGameScene();
}